import contextlib
import gc
import shutil
import tempfile
import typing
from pathlib import Path
from typing import Literal

import pydantic
import pytest

from .io import TimingsOutputFile
from .mocks import MockRecorder
from .data import (
    CollectionTiming,
    FixtureTiming,
    TestTiming,
    GCTiming,
)
from .utils import is_generator_fixture
from .timer import Timer, measure_time

if typing.TYPE_CHECKING:
    from _pytest.fixtures import FixtureDef, SubRequest


@pytest.hookimpl
def pytest_addoption(parser: pytest.Parser):
    group = parser.getgroup("scrutinize", "Structured timing output")
    group.addoption(
        "--scrutinize",
        metavar="N",
        action="store",
        type=Path,
        help="Store structured timing output to this file",
    )
    group.addoption(
        "--scrutinize-func",
        action="append",
        type=str,
        nargs="?",
        help="Comma separated list of functions to record",
    )
    group.addoption(
        "--scrutinize-gc", action="store_true", help="Record garbage collections"
    )
    group.addoption(
        "--scrutinize-django-sql",
        nargs="?",
        choices=["hash", "query"],
        default=False,
        const=True,
        help="Record Django SQL queries",
    )


class Config(pydantic.BaseModel):
    output_path: Path
    mocks: frozenset[str]
    enable_gc: bool
    enable_django_sql: Literal[True, "query"] | None


def pytest_configure(config: pytest.Config):
    if output_path := config.getoption("--scrutinize"):
        assert isinstance(output_path, Path)

        enable_gc = typing.cast(bool, config.getoption("--scrutinize-gc") or False)

        enable_django_sql = typing.cast(
            Literal[True, "query"] | None,
            config.getoption("--scrutinize-django-sql") or None,
        )

        mocks = typing.cast(list[str], config.getoption("--scrutinize-func"))
        if mocks is None:
            mocks = frozenset()
        else:
            mocks = {
                stripped_mock
                for mocks_arg in mocks
                for mock_path in mocks_arg.split(",")
                if (stripped_mock := mock_path.strip())
            }
        plugin_config = Config(
            output_path=output_path,
            mocks=frozenset(mocks),
            enable_gc=enable_gc,
            enable_django_sql=enable_django_sql,
        )

        plugin_cls: type[DetailedTimingsPlugin]
        has_xdist = config.pluginmanager.get_plugin("xdist") is not None

        from . import plugin_xdist

        is_xdist_master = plugin_xdist.is_master()
        match (has_xdist, is_xdist_master):
            case (True, True):
                plugin_cls = plugin_xdist.XDistMasterDetailedTimingsPlugin
            case (True, False):
                plugin_cls = plugin_xdist.XDistWorkerDetailedTimingsPlugin
            case (False, _):
                plugin_cls = DetailedTimingsPlugin
            case _:
                assert False, f"unreachable: {has_xdist=} {is_xdist_master=}"

        plugin = plugin_cls(plugin_config)
        config.pluginmanager.register(plugin, name=__name__)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtestloop(session: pytest.Session):
    if plugin := session.config.pluginmanager.get_plugin(__name__):
        with plugin.run(session):
            yield
    else:
        yield


class DetailedTimingsPlugin:
    config: Config
    output: TimingsOutputFile
    mock_recorder: MockRecorder

    def __init__(self, config: Config):
        self.config = config

        temp_dir = Path(tempfile.mkdtemp())
        temp_output_path = temp_dir / "output.jsonl.gz"
        self.output = TimingsOutputFile(temp_output_path)
        self.mock_recorder = MockRecorder(
            mocks=config.mocks,
            output=self.output,
            enable_django_sql=self.config.enable_django_sql,
        )

        if config.enable_gc:
            self.setup_gc_callbacks()

    @contextlib.contextmanager
    def run(self, session: pytest.Session) -> typing.Generator[typing.Self, None, None]:
        with self.output.initialize_writer(), self.mock_recorder.initialize_mocks():
            yield self

        self.create_final_output_file(session)

    def create_final_output_file(self, session: pytest.Session):
        shutil.move(src=self.output.path, dst=self.config.output_path)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_collection(self, session: pytest.Session):
        with measure_time() as timer:
            yield

        self.output.add_timing(CollectionTiming(runtime=timer.elapsed))

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item: pytest.Item, nextitem: pytest.Item | None):
        try:
            yield
        finally:
            self.output.flush_buffer()

    @pytest.hookimpl(hookwrapper=True)
    def pytest_pyfunc_call(self, pyfuncitem: pytest.Function):
        with self.mock_recorder.record(test_id=pyfuncitem.nodeid, fixture_name=None):
            with measure_time() as timer:
                yield

        test_timing = TestTiming(
            name=pyfuncitem.name,
            test_id=pyfuncitem.nodeid,
            requires=pyfuncitem.fixturenames,
            runtime=timer.elapsed,
        )

        self.output.add_timing(test_timing)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_fixture_setup(self, fixturedef: "FixtureDef", request: "SubRequest"):
        is_function_scope = fixturedef.scope == "function"
        full_name = f"{fixturedef.func.__module__}.{fixturedef.func.__qualname__}"

        # Don't associate non-function scoped fixtures with a given test
        test_id = request.node.nodeid if is_function_scope else None

        setup_timer: Timer
        teardown_timer: Timer | None = None

        def fixture_done():
            nonlocal setup_timer, teardown_timer
            teardown_duration_ns = (
                teardown_timer.elapsed if teardown_timer is not None else None
            )

            self.output.add_timing(
                FixtureTiming(
                    name=full_name,
                    short_name=fixturedef.func.__qualname__,
                    test_id=test_id,
                    scope=request.scope,
                    setup=setup_timer.elapsed,
                    teardown=teardown_duration_ns,
                )
            )

        fixturedef.addfinalizer(fixture_done)

        if not is_generator_fixture(fixturedef.func):
            with self.mock_recorder.record(test_id=test_id, fixture_name=full_name):
                with measure_time() as setup_timer:
                    yield
        else:
            # We want to capture the teardown times for fixtures. This is non-trivial as
            # pytest exposes no hooks to allow you to do this.
            # However, we can use finalizers for this: we first attach a finalizer _before_
            # the fixture is executed, then attach a finalizer _after_ the fixture is executed.
            # Pytest hooks are run in reverse order: the first hook to run
            # will be the `record_teardown_finish` finalizer, and the first will be the `record_teardown_start`.

            teardown_mock_capture: typing.ContextManager | None = None
            teardown_timer = Timer()

            def teardown_fixture_start():
                nonlocal teardown_mock_capture
                teardown_mock_capture = self.mock_recorder.record(
                    test_id=test_id, fixture_name=full_name
                )
                teardown_mock_capture.__enter__()
                teardown_timer.__enter__()

            def teardown_fixture_finish():
                teardown_timer.__exit__(None, None, None)
                if teardown_mock_capture is not None:
                    teardown_mock_capture.__exit__(None, None, None)

            fixturedef.addfinalizer(teardown_fixture_finish)
            with self.mock_recorder.record(test_id=test_id, fixture_name=full_name):
                with measure_time() as setup_timer:
                    yield

            fixturedef.addfinalizer(teardown_fixture_start)

    def setup_gc_callbacks(self):
        gc_timer = Timer()

        def gc_callback(phase: Literal["start", "stop"], info: dict[str, int]):
            if phase == "start":
                gc_timer.start()
            else:
                gc_timer.stop()

                self.output.add_timing(
                    GCTiming(
                        runtime=gc_timer.elapsed,
                        collected_count=info["collected"],
                        generation=info["generation"],
                    )
                )

        gc.callbacks.append(gc_callback)



import gzip
from pathlib import Path
from typing import Callable

import pytest
from _pytest.pytester import RunResult, Pytester

from pytest_scrutinize import Timing, TimingAdapter

pytest_plugins = [
    "pytester",
]


def read_results_file(path: Path) -> list[Timing]:
    with gzip.open(path, mode="rt") as fd:
        return [TimingAdapter.validate_json(line) for line in fd]


@pytest.fixture(params=[True, False], ids=["xdist", "normal"])
def with_xdist(request) -> bool:
    return request.param


@pytest.fixture()
def output_file(tmp_path) -> Path:
    return tmp_path / "output.jsonl.gz"


@pytest.fixture()
def run_tests(
    pytester_pretty: Pytester, with_xdist, output_file
) -> Callable[..., tuple[RunResult, list[Timing]]]:
    flags = []
    if with_xdist:
        flags.append("-n 2")

    def _run(test_name: str, *args: str):
        pytester_pretty.copy_example(test_name)
        result: RunResult = pytester_pretty.runpytest(
            "--scrutinize", output_file, *flags, *args
        )
        parsed_results = read_results_file(output_file)
        return result, parsed_results

    return _run



import contextlib
import dataclasses
import time
from datetime import datetime, UTC, timedelta
from typing import Generator

import pydantic
from pydantic import computed_field
from types import SimpleNamespace

# Freezegun (and likely other similar libraries) does some insane stuff to try and ensure that
# time.* and datetime.* functions are patched - it scans every module and looks for any module
# level attribute that is a function it is patching, then patches it.
# We store the original functions inside a class so that it is not replaced under our feet.

_time_funcs = SimpleNamespace(perf_ns=time.perf_counter_ns, now=datetime.now)


def now() -> datetime:
    return _time_funcs.now(tz=UTC)


@contextlib.contextmanager
def measure_time() -> Generator["Timer", None, None]:
    with Timer() as timer:
        yield timer


@dataclasses.dataclass
class Timer:
    _start: int | None = None
    _end: int | None = None

    @property
    def elapsed(self) -> "Duration":
        if self._start is None:
            raise RuntimeError("Timer not started")

        end = _time_funcs.perf_ns() if self._end is None else self._end
        return Duration(as_nanoseconds=end - self._start)

    def start(self):
        self.reset()
        self._start = _time_funcs.perf_ns()

    def stop(self):
        self._end = _time_funcs.perf_ns()

    def reset(self):
        self._start, self._end = None, None

    def __enter__(self) -> "Timer":
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class Duration(pydantic.BaseModel):
    as_nanoseconds: int

    @computed_field  # type: ignore[prop-decorator]
    @property
    def as_microseconds(self) -> int:
        return self.as_nanoseconds // 1_000

    @computed_field  # type: ignore[prop-decorator]
    @property
    def as_iso(self) -> timedelta:
        return timedelta(microseconds=self.as_microseconds)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def as_text(self) -> str:
        return f"{self.as_microseconds} microseconds"

    def __add__(self, other: "Duration") -> "Duration":
        return Duration(as_nanoseconds=self.as_nanoseconds + other.as_nanoseconds)



import collections
import typing
from typing import Type, Hashable

import pytest
from _pytest.pytester import RunResult

from pytest_scrutinize import (
    Timing,
    CollectionTiming,
    WorkerTiming,
    TestTiming as PyTestTiming,
    FixtureTiming,
    MockTiming,
    DjangoSQLTiming,
    GCTiming,
)
from pytest_scrutinize.timer import Duration

T = typing.TypeVar("T", bound=Timing)


def get_timing_items(results: list[Timing], cls: Type[T]) -> list[T]:
    return [result for result in results if isinstance(result, cls)]


def assert_duration(duration: Duration | None):
    assert duration is not None
    assert duration.as_nanoseconds != 0
    assert duration.as_microseconds == duration.as_iso.microseconds


H = typing.TypeVar("H", bound=Hashable)


def assert_unique(items: typing.Iterable[H]) -> set[H]:
    items = list(items)
    unique = set(items)
    assert len(unique) == len(items)
    return unique


def assert_not_master(results: list[T]):
    for result in results:
        assert result.meta.worker != "master"


def assert_results_collection(results: list[Timing]):
    collection_times = get_timing_items(results, CollectionTiming)
    assert collection_times != []
    unique = assert_unique(ev.meta.worker for ev in collection_times)
    assert "master" in unique

    for collection in collection_times:
        assert_duration(collection.runtime)


def assert_xdist_workers_ready(results: list[Timing]):
    worker_ready = get_timing_items(results, WorkerTiming)
    assert worker_ready != []
    assert_unique(ev.meta.worker for ev in worker_ready)
    assert_not_master(worker_ready)
    for ev in worker_ready:
        assert_duration(ev.ready)
        assert_duration(ev.runtime)


def assert_tests(results: list[Timing], is_xdist: bool):
    test_timings = get_timing_items(results, PyTestTiming)
    assert test_timings != []
    assert_unique(ev.test_id for ev in test_timings)

    if is_xdist:
        assert_not_master(test_timings)

    for test_timing in test_timings:
        assert_duration(test_timing.runtime)


def assert_fixture(fixture: FixtureTiming, root_name: str):
    assert_duration(fixture.setup)

    if not fixture.name.startswith(f"{root_name}."):
        return

    if fixture.name.startswith(f"{root_name}.teardown_"):
        assert_duration(fixture.teardown)
    else:
        assert fixture.teardown is None


def assert_fixtures(results: list[Timing], is_xdist: bool, root_name: str):
    fixture_timings = get_timing_items(results, FixtureTiming)
    assert fixture_timings != []
    if is_xdist:
        assert_not_master(fixture_timings)

    test_names_map = {
        test.test_id: test for test in get_timing_items(results, PyTestTiming)
    }

    for fixture_timing in fixture_timings:
        if fixture_timing.test_id is not None:
            assert fixture_timing.test_id in test_names_map
        assert_fixture(fixture_timing, root_name)


def assert_mocks(results: list[Timing], is_xdist: bool, root_name: str):
    fixture_timings = get_timing_items(results, FixtureTiming)
    mock_timings = get_timing_items(results, MockTiming)
    assert mock_timings != []
    if is_xdist:
        assert_not_master(mock_timings)

    mock_names = {mock.name for mock in mock_timings}
    assert mock_names == {
        "urllib.parse.parse_qs",
        "urllib.parse.quote",
        "urllib.parse.urlparse",
    }

    assert set(
        mock.test_id.split("::")[1] for mock in mock_timings if mock.test_id
    ) == {"test_case"}

    fixture_map = {fixture.name: fixture for fixture in fixture_timings}

    # All fixtures should have called the mock, excluding the indirect fixture
    expected_fixtures = {
        f"{root_name}.fixture",
        f"{root_name}.teardown_fixture",
    }
    fixtures_calling_mock = {
        mock_timing.fixture_name
        for mock_timing in mock_timings
        if mock_timing.fixture_name is not None
    }
    assert fixtures_calling_mock == expected_fixtures

    for mock_timing in mock_timings:
        assert_duration(mock_timing.runtime)
        assert mock_timing.fixture_name or mock_timing.test_id
        if mock_timing.test_id:
            assert mock_timing.test_id == f"{root_name}.py::test_case"
        # Check that the fixture name is in the map
        if mock_timing.fixture_name is not None:
            assert mock_timing.fixture_name in fixture_map


def assert_suite(result: RunResult, timings: list[Timing], with_xdist: bool):
    result.assert_outcomes(passed=1)

    assert_results_collection(timings)
    assert_tests(timings, with_xdist)
    assert_fixtures(timings, with_xdist, root_name="test_suite")

    if with_xdist:
        assert_xdist_workers_ready(timings)


def test_simple(run_tests, output_file, with_xdist):
    result, timings = run_tests("test_simple.py")
    assert_suite(result, timings, with_xdist)


def test_mocks(run_tests, output_file, with_xdist):
    result, timings = run_tests(
        "test_mock.py",
        "--scrutinize-func=urllib.parse.urlparse,urllib.parse.parse_qs",
        "--scrutinize-func=urllib.parse.quote",
    )
    assert_suite(result, timings, with_xdist)
    assert_mocks(timings, with_xdist, root_name="test_mock")


def test_gc(run_tests, output_file, with_xdist):
    result, timings = run_tests("test_simple.py", "--scrutinize-gc")
    assert_suite(result, timings, with_xdist)
    gc_timings = get_timing_items(timings, GCTiming)
    assert gc_timings != []
    for gc_timing in gc_timings:
        assert_duration(gc_timing.runtime)


@pytest.mark.parametrize("with_query", [True, False])
def test_django(run_tests, output_file, with_xdist, with_query):
    flag = "--scrutinize-django-sql"
    if with_query:
        flag = f"{flag}=query"
    result, timings = run_tests(
        "test_django.py", "--ds=tests.django_app.settings", flag
    )
    assert_suite(result, timings, with_xdist)
    sql_timings = get_timing_items(timings, DjangoSQLTiming)
    assert sql_timings != []

    timings_by_fixture: dict[str, list[DjangoSQLTiming]] = collections.defaultdict(list)
    timings_by_test: dict[str, list[DjangoSQLTiming]] = collections.defaultdict(list)

    for sql_timing in sql_timings:
        assert_duration(sql_timing.runtime)
        if sql_timing.fixture_name:
            timings_by_fixture[sql_timing.fixture_name].append(sql_timing)
        elif sql_timing.test_id:
            # Only test queries, no fixtures at all
            timings_by_test[sql_timing.test_id].append(sql_timing)

    # All of our SQL hashes (and queries) should be the same
    sql_hashes = set(
        (timing.sql_hash, timing.sql)
        for fixture_name in ("test_django.teardown_fixture", "test_django.fixture")
        for timing in timings_by_fixture[fixture_name]
    )
    assert len(sql_hashes) == 1

    assert_unique(
        (timing.sql_hash, timing.sql)
        for timing in timings_by_test["test_django.py::test_case"]
    )


def test_all(run_tests, output_file, with_xdist):
    result, timings = run_tests(
        "test_simple.py",
        "--scrutinize-django-sql",
        "--scrutinize-gc",
        "--scrutinize-func=urllib.parse.urlparse",
    )
    assert_suite(result, timings, with_xdist)



import pytest
from urllib import parse

url = "https://google.com/foobar"


def call_mocked_function() -> int:
    assert parse.urlparse(url).path == "/foobar"
    assert parse.quote("foo") == "foo"
    assert parse.parse_qs("foo=bar") == {"foo": ["bar"]}
    return 123


@pytest.fixture()
def indirect_fixture():
    call_mocked_function()


@pytest.fixture()
def teardown_fixture():
    call_mocked_function()
    yield
    call_mocked_function()


@pytest.fixture()
def fixture(request: pytest.FixtureRequest):
    request.getfixturevalue("indirect_fixture")
    call_mocked_function()
    return


def test_case(teardown_fixture, fixture):
    call_mocked_function()
    assert True



INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "tests.django_app.apps.DjangoAppConfig",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}



import typing

import pydantic
from typing import Union
from .data import (
    GCTiming,
    CollectionTiming,
    WorkerTiming,
    MockTiming,
    TestTiming,
    FixtureTiming,
    DjangoSQLTiming,
)

Timing = typing.Annotated[
    Union[
        GCTiming,
        CollectionTiming,
        WorkerTiming,
        MockTiming,
        TestTiming,
        FixtureTiming,
        DjangoSQLTiming,
    ],
    pydantic.Field(discriminator="type"),
]

TimingAdapter = pydantic.TypeAdapter(Timing)



import os
import shutil
import typing

from pathlib import Path
from typing import Any, Sequence

import pytest

from pytest_scrutinize.plugin import DetailedTimingsPlugin
from pytest_scrutinize.data import WorkerTiming, Meta

from pytest_scrutinize.timer import Timer
from .io import TimingsOutputFile

if typing.TYPE_CHECKING:
    from .plugin import Config

    try:
        from xdist.workermanage import WorkerController
    except ImportError:
        pass


def get_worker_id() -> str:
    return os.environ.get("PYTEST_XDIST_WORKER", "master")


def is_master() -> bool:
    return get_worker_id() == "master"


_worker_output_key = f"{__name__}.output"


class XDistWorkerDetailedTimingsPlugin(DetailedTimingsPlugin):
    @pytest.hookimpl()
    def pytest_sessionfinish(self, session: pytest.Session, exitstatus: int):
        if workeroutput := getattr(session.config, "workeroutput", None):
            workeroutput[_worker_output_key] = str(self.output.path.absolute())

    def create_final_output_file(self, session: pytest.Session):
        return


class XDistMasterDetailedTimingsPlugin(DetailedTimingsPlugin):
    setup_nodes_timer: Timer
    worker_timings: dict[str, WorkerTiming]
    worker_output_files: list[Path]

    def __init__(self, config: "Config"):
        super().__init__(config=config)

        self.setup_nodes_timer = Timer()
        self.worker_timings = {}
        self.worker_output_files = []

    @pytest.hookimpl()
    def pytest_xdist_setupnodes(self, config: pytest.Config, specs: Sequence[Any]):
        self.setup_nodes_timer.start()

    @pytest.hookimpl()
    def pytest_testnodeready(self, node: "WorkerController"):
        duration = self.setup_nodes_timer.elapsed

        worker_id = node.workerinfo["id"]
        self.worker_timings[worker_id] = WorkerTiming(
            meta=Meta(worker=worker_id),
            ready=duration,
        )

    def create_final_output_file(self, session: pytest.Session):
        final_output_file = TimingsOutputFile(path=self.config.output_path)
        files_to_combine = [self.output.path] + self.worker_output_files

        with final_output_file.initialize_writer() as output_writer:
            for input_path in files_to_combine:
                with TimingsOutputFile(path=input_path).get_reader() as output_reader:
                    shutil.copyfileobj(fsrc=output_reader, fdst=output_writer)

    def pytest_testnodedown(self, node: "WorkerController", error: Any):
        if workeroutput := getattr(node, "workeroutput", None):
            worker_id = node.workerinfo["id"]
            if worker_timing := self.worker_timings.get(worker_id, None):
                # time since setup nodes was invoked
                worker_timing.runtime = self.setup_nodes_timer.elapsed
                self.output.add_timing(worker_timing)

            if output_path := workeroutput.get(_worker_output_key, None):
                self.worker_output_files.append(Path(output_path))



import abc
import threading
from datetime import datetime
from typing import Literal

import pydantic
from pydantic import computed_field
from pydantic.fields import Field

from .timer import now, Duration


def get_worker_field_default() -> str:
    # Work around circular imports. to-do: refactor this
    from .plugin_xdist import get_worker_id

    return get_worker_id()


class Meta(pydantic.BaseModel):
    worker: str = Field(default_factory=get_worker_field_default)
    recorded_at: datetime = Field(default_factory=now)

    thread_id: int = Field(default_factory=threading.get_native_id)
    thread_name: str = Field(default_factory=lambda: threading.current_thread().name)


class BaseTiming(pydantic.BaseModel, abc.ABC):
    meta: Meta = Field(default_factory=Meta)


class GCTiming(BaseTiming):
    type: Literal["gc"] = "gc"

    runtime: Duration
    collected_count: int
    generation: int


class CollectionTiming(BaseTiming):
    type: Literal["collection"] = "collection"

    runtime: Duration


class WorkerTiming(BaseTiming):
    type: Literal["worker"] = "worker"

    ready: Duration
    runtime: Duration | None = None


class BaseMockTiming(BaseTiming, abc.ABC):
    name: str
    test_id: str | None
    fixture_name: str | None

    runtime: Duration


class MockTiming(BaseMockTiming):
    type: Literal["mock"] = "mock"


class DjangoSQLTiming(BaseMockTiming):
    type: Literal["django-sql"] = "django-sql"
    sql_hash: str
    sql: str | None


class TestTiming(BaseTiming):
    type: Literal["test"] = "test"

    name: str
    test_id: str
    requires: list[str]

    runtime: Duration


class FixtureTiming(BaseTiming):
    type: Literal["fixture"] = "fixture"

    name: str
    short_name: str
    test_id: str | None
    scope: str

    setup: Duration
    teardown: Duration | None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def runtime(self) -> Duration:
        if self.teardown is not None:
            return self.setup + self.teardown
        return self.setup



import gc

import pytest


@pytest.fixture()
def teardown_fixture():
    yield


@pytest.fixture()
def fixture():
    return


def test_case(teardown_fixture, fixture):
    # Force a collection, for --scrutinize-gc tests
    gc.collect()
    assert True



import contextlib
import pkgutil
from dataclasses import dataclass, field
from typing import Any, Callable, Self, Literal
from unittest import mock
import hashlib
import pydantic

from pytest_scrutinize.io import TimingsOutputFile
from pytest_scrutinize.timer import measure_time, Duration
from pytest_scrutinize.data import MockTiming, DjangoSQLTiming, BaseMockTiming


class SingleMockRecorder(pydantic.BaseModel):
    name: str
    mocked: Any
    original_callable: Callable

    @classmethod
    def from_dotted_path(cls, name: str, mock_path: str, **kwargs) -> Self:
        class_path, attribute_name = mock_path.rsplit(".", 1)
        original_object = pkgutil.resolve_name(class_path)
        original_callable = getattr(original_object, attribute_name)
        mocked = mock.patch(mock_path, side_effect=None, autospec=True)
        return cls(
            name=name, mocked=mocked, original_callable=original_callable, **kwargs
        )

    def record_timing(
        self,
        fixture_name: str | None,
        elapsed: Duration,
        test_id: str | None,
        *,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> BaseMockTiming:
        return MockTiming(
            fixture_name=fixture_name,
            runtime=elapsed,
            name=self.name,
            test_id=test_id,
        )

    @contextlib.contextmanager
    def record_mock(
        self, output: TimingsOutputFile, test_id: str | None, fixture_name: str | None
    ):
        if self.mocked.kwargs["side_effect"] is not None:
            raise RuntimeError(f"Recursive mock call for mock {self}")

        def wrapped(*args, **kwargs):
            with measure_time() as timer:
                result = self.original_callable(*args, **kwargs)
            output.add_timing(
                self.record_timing(
                    fixture_name, timer.elapsed, test_id, args=args, kwargs=kwargs
                )
            )
            return result

        self.mocked.kwargs["side_effect"] = wrapped
        try:
            with self.mocked:
                yield
        finally:
            self.mocked.kwargs["side_effect"] = None


class DjangoSQLRecorder(SingleMockRecorder):
    mode: Literal[True, "query"]

    def record_timing(
        self,
        fixture_name: str | None,
        elapsed: Duration,
        test_id: str | None,
        *,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> DjangoSQLTiming:
        # The django.db.backends.utils.CursorWrapper._execute function takes the
        # SQL as the second argument (the first being `self`):
        query: str = args[1]
        sql_hash = hashlib.sha256(query.encode(), usedforsecurity=False).hexdigest()

        sql = None
        if self.mode == "query":
            # Hash it:
            sql = query
        return DjangoSQLTiming(
            fixture_name=fixture_name,
            runtime=elapsed,
            name=self.name,
            test_id=test_id,
            sql_hash=sql_hash,
            sql=sql,
        )


@dataclass
class MockRecorder:
    mocks: frozenset[str]
    output: TimingsOutputFile
    enable_django_sql: Literal[True, "query"] | None

    _mock_funcs: dict[str, SingleMockRecorder] = field(default_factory=dict)

    @contextlib.contextmanager
    def record(self, test_id: str | None, fixture_name: str | None):
        # We want to avoid recursive mock calls, which can happen when using `getfixturevalue`.
        # Detecting if something is a mock is a pain, so we just use a token-style system:
        # we take the entire _mock_funcs dictionary when we set up the mocks, and replace it
        # when we are done.

        if not self._mock_funcs:
            yield
            return

        mock_funcs = self._mock_funcs
        self._mock_funcs = {}

        try:
            with contextlib.ExitStack() as stack:
                for single_mock in mock_funcs.values():
                    stack.enter_context(
                        single_mock.record_mock(self.output, test_id, fixture_name)
                    )
                yield
        finally:
            self._mock_funcs = mock_funcs

    @contextlib.contextmanager
    def initialize_mocks(self):
        for mock_path in self.mocks:
            self._mock_funcs[mock_path] = SingleMockRecorder.from_dotted_path(
                name=mock_path, mock_path=mock_path
            )

        if self.enable_django_sql is not None:
            self._mock_funcs["django_sql"] = DjangoSQLRecorder.from_dotted_path(
                name="django_sql",
                mock_path="django.db.backends.utils.CursorWrapper._execute",
                mode=self.enable_django_sql,
            )

        try:
            yield
        finally:
            self._mock_funcs.clear()



import contextlib
import gzip
import typing
from dataclasses import dataclass, field
from pathlib import Path

if typing.TYPE_CHECKING:
    from pytest_scrutinize.data import BaseTiming


@dataclass
class TimingsOutputFile:
    path: Path
    buffer: list["BaseTiming"] = field(default_factory=list)

    fd: typing.TextIO | None = None

    def add_timing(self, timing: "BaseTiming"):
        self.buffer.append(timing)

    def flush_buffer(self):
        if self.fd is None:
            raise RuntimeError("Output file not opened")
        # Get a reference to the buffer, then replace it with an
        # empty list. We do this because the GC callbacks _could_
        # cause an append to the list mid-iteration, or after the
        # write loop has finished.
        buffer = self.buffer
        self.buffer = []
        for timing in buffer:
            self.fd.write(timing.model_dump_json())
            self.fd.write("\n")

    @contextlib.contextmanager
    def initialize_writer(self) -> typing.Generator[typing.TextIO, None, None]:
        if self.fd is not None:
            raise RuntimeError("Output file already opened")

        with gzip.open(self.path, mode="wt", compresslevel=1) as fd:
            self.fd = fd
            try:
                yield fd
            finally:
                self.flush_buffer()
                self.fd = None

    @contextlib.contextmanager
    def get_reader(self) -> typing.Generator[typing.TextIO, None, None]:
        if self.fd is not None:
            raise RuntimeError("Output file not closed")

        with gzip.open(self.path, mode="rt") as fd:
            yield fd



import pytest
from tests.django_app.models import DummyModel


@pytest.fixture()
def teardown_fixture():
    DummyModel.objects.create(foo="foobar")
    yield
    DummyModel.objects.create(foo="foobar")


@pytest.fixture()
def fixture():
    DummyModel.objects.create(foo="foobar")


@pytest.mark.django_db
def test_case(teardown_fixture, fixture):
    obj = DummyModel.objects.create(foo="foobar")
    assert DummyModel.objects.count() == 3
    assert len(list(DummyModel.objects.all())) == 3
    obj.delete()
